clc;
clear;
close all;

%% plot the objective function

% define the function as anonymous function handle
% you can then evaluate it as f(x,y) (elementwise)
f = @(x,y) .5*( (x-1).^2 + 100*(y-x.^2).^2 + y.^2);

% Create circle grid of x y values 
R=0:.002:sqrt(2);
TH=2*pi*(0:.002:1); 
X=R'*cos(TH); 
Y=R'*sin(TH);

% x = linspace(-1,1,200);
% [X,Y] = meshgrid(x,x);

% calculate objective function values
% add 1 and take log for better visualization
Z=log(1 + f(X,Y));
% Z=f(X,Y);
figure(1); clf;
colormap('gray');

% subplot 1 with surface plot and contour plot
subplot(1,2,1); hold on
view([124 34]);
grid('on'); 
axis square;
surf(X,Y,Z,'LineStyle','none');
contour(X,Y,Z);
title('Surface and contour plot of the nonlinear function');
xlabel('x')
ylabel('y')
zlabel('log(1+f(x,y))')

% subplot 2 with contour plot only
subplot(1,2,2); hold on
grid('on'); axis square;
contour(X,Y,Z, 'Displayname', '$f(x,y)$');
% read current axis limits and fix them at these values
ax = axis(); axis(ax);


%% grad and hessians

% gradient
fgrad = @(x,y) [x-1-200*x.*y+200*x.^3;
                101*y - 100* x.^2 ];

% true hessian
fhess = @(x,y) [1-200*y+600*x^2, -200*x;
                -200*x, 101];

% Gauss newton hessian
B_GN = @(x,y) [1+400*x^2, -200*x;
               -200*x   , 101];
            
z0 = [-1;1];                            % initial guesss
tol = 1e-3;                             % tolerance for stopping criterion
maxit = 5000;
%% newton Type algorithms
% introduce z = [x;y]
fgradz = @(z) fgrad(z(1), z(2));
% exact hessian
tic
[Z_ex, grad_ex] = newtonType(fgradz, @(z) fhess(z(1), z(2)), z0, tol, maxit);
t_ex = toc;
% Gauss Newton
tic
[Z_GN, grad_GN] = newtonType(fgradz, @(z) B_GN(z(1), z(2)), z0, tol, maxit);
t_GN = toc;
% three versions of gradient descent
tic
[Z_gd1, grad_gd1] = newtonType(fgradz, @(z) 100*eye(2), z0, tol, maxit);
t_gd1 = toc;
tic
[Z_gd2, grad_gd2] = newtonType(fgradz, @(z) 200*eye(2), z0, tol, maxit);
t_gd2 = toc;
tic
[Z_gd5, grad_gd5] = newtonType(fgradz, @(z) 500*eye(2), z0, tol, maxit);
t_gd5 = toc;

%% print infos
fprintf('\n-----------------------------------\n')
fprintf('Method\t\t  #iter\t  t in s\n')
fprintf('-----------------------------------\n')
fprintf('exact hessian      %4d   %.3e\n', size(Z_ex,2)-1, t_ex)
fprintf('Gauss-Newton       %4d   %.3e\n', size(Z_GN,2)-1, t_GN)
fprintf('grad desc, a=100   %4d   %.3e\n', size(Z_gd1,2)-1, t_gd1)
fprintf('grad desc, a=200   %4d   %.3e\n', size(Z_gd2,2)-1, t_gd2)
fprintf('grad desc, a=500   %4d   %.3e\n', size(Z_gd5,2)-1, t_gd5)

%% add iterates to plot
figure(1); subplot(1,2,2);
x = linspace(ax(1),ax(2),100);
plot(x, x.^2,'k--', 'Displayname', '$x=y^2$')
plot(Z_ex(1,:), Z_ex(2,:), 'b-x', 'Displayname', 'exact Hessian')
plot(Z_GN(1,:), Z_GN(2,:), 'r-x', 'Displayname', 'GN')
plot(Z_gd5(1,:), Z_gd5(2,:), 'g-x', 'Displayname', 'GD, $\alpha=500$')
plot(Z_gd2(1,:), Z_gd2(2,:), 'k-x', 'Displayname', 'GD, $\alpha=200$')
plot(Z_gd1(1,:), Z_gd1(2,:), 'c-x', 'Displayname', 'GD, $\alpha=100$')
leg = legend('location', 'southeast');
set(leg,'interpreter','latex')

%%
figure
axis square;
contour(X,Y,Z, 'Displayname', '$f(x,y)$','LineWidth',2);
colormap( flipud(gray(256)) )
xlabel('$x$','Interpreter','latex')
ylabel('$y$','Interpreter','latex')
t_sim = 5;
% read current axis limits and fix them at these values
ax = axis(); axis(ax);
grid on
hold on
plot(x, x.^2,'k--', 'Displayname', '$x=y^2$')

% EH
plot(Z_ex(1,1), Z_ex(2,1), 'b-x','LineWidth',2, 'Displayname', 'exact Hessian');
for ii = 2:length(Z_ex(1,:))
    plot(Z_ex(1,1:ii), Z_ex(2,1:ii), 'b-x','LineWidth',2)
    pause(t_sim/length(Z_ex(1,:)))
end

%GN
plot(Z_GN(1,1), Z_GN(2,1), 'r-x', 'Displayname', 'GN','LineWidth',2);
for ii = 2:length(Z_GN(1,:))
    plot(Z_GN(1,1:ii), Z_GN(2,1:ii), 'r-x','LineWidth',2)
    pause(t_sim/length(Z_GN(1,:)))
end

plot(Z_gd5(1,1), Z_gd5(2,1), 'g-x', 'Displayname', 'GD, $\alpha=500$')
for ii = 2:5:length(Z_gd5(1,:))
    plot(Z_gd5(1,ii-1:ii), Z_gd5(2,ii-1:ii), 'g-x','LineWidth',2)
    pause(t_sim/length(Z_gd5(1,:)))
end


plot(Z_gd2(1,1), Z_gd2(2,1), 'k-x', 'Displayname', 'GD, $\alpha=200$')
for ii = 2:3:length(Z_gd2(1,:))
    plot(Z_gd2(1,ii-1:ii), Z_gd2(2,ii-1:ii), 'k-x','LineWidth',2)
    pause(t_sim/length(Z_gd2(1,:)))
end

%
plot(Z_gd1(1,1), Z_gd1(2,1), 'c-x', 'Displayname', 'GD, $\alpha=100$')
for ii = 2:10:length(Z_gd1(1,:))
    plot(Z_gd1(1,ii-1:ii), Z_gd1(2,ii-1:ii), 'c-x','LineWidth',2)
    pause(t_sim/length(Z_gd1(1,:)))
end
leg = legend('location', 'southeast');
set(leg,'interpreter','latex')


