import matplotlib.pyplot as plt
import matplotlib.animation as animation
import numpy as np


def animate_chain(x_traj, dx_traj):
    fig = plt.figure(1)
    ax1 = plt.subplot(2,1,1)
    
    ax1.set_xlim([-2, 2])
    ax1.set_ylim([-3, 1])
    ax1.set_title(r'Position of chain at current iterate')
    ax1.set_xlabel(r'$y$')
    ax1.set_ylabel(r'$z$')

    ax2 = plt.subplot(2,1,2)
    ax2.set_xlim([0, np.size(x_traj[0])])
    ax2.set_ylim([-1, 1])
    ax2.grid(axis='y')
    ax2.set_title(r'full step of each optimization variable ($dx_i$)')
    ax2.set_xlabel(r'$i$')
    ax2.set_ylabel(r'$|\Delta x_i|$')

    # set a proper step that there are roughly 100 frames
    max_iter = len(x_traj)
    if max_iter > 100:
        step = max_iter // 100
        digits = int(np.log10(step))
        step = step // 10**digits * 10**digits
    else:
        step = 1

    all_artist = []
    for i in range(0, max_iter, step):
        art0 = ax1.text(1, -2, f"iter={i}")
        x = x_traj[i]
        y = x[0:None:2]
        z = x[1:None:2]
        [art1] = ax1.plot(y, z, 'b--')
        [art2] = ax1.plot(y, z, 'ro')
        dx = dx_traj[i]
        [art3] = ax2.plot(dx, 'k.')
        all_artist.append([art0, art1, art2, art3])

    return animation.ArtistAnimation(fig, all_artist, interval=100, repeat=True, repeat_delay=1500, blit=True)
